/*
 *    Rosegarden MIDI Sequencer
 *
 *    File:           Record_LINUX.c
 *
 *    Description:    Sequencer Record functions.
 *
 *    Author:         AJG
 *
 *
 *    Description:    Sequencer Record functions.
 *
 *
 *    History:
 *
 *    Update    Date            Programmer      Comments
 *    ======    ====            ==========      ========
 *    001       21/04/94        AJG             File Created.
 *    002       17/02/95        cc              Added #ifdefs for HAVE_Z8530
 *    003       24/05/95        cc              Archer Sully's SGI MIDI lib
 *                                                                       stuff
 *    004       10/05/96        rwb             Linuxified.
 *
 *
 *
 */

#undef POSIX_PLEASE
#undef _POSIX_SOURCE

#include <sys/soundcard.h>
#include <sys/ioctl.h>
#include <errno.h>

#include "Record.h"
#include "Sequence.h"

#include "Globals.h"
#include "Consts.h"

#include <MidiXInclude.h>
#include <MidiFile.h>
#include <MidiErrorHandler.h>
#include <Debug.h>
#include <MidiBHeap.h>
#include <MidiTrack.h>

#include "TrackList.h"

#include <signal.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/times.h>
#include <unistd.h>
#include <sys/signal.h>

long	 RecordTempo         = 500000;
int      MidiEventBufferSize = 10000;

extern char *MidiPortName;
EventList Midi_SeqConvertRawDataToEventList(void);

typedef struct
{
	clock_t	Time;
	byte	Bytes[3];
}
MIDIRawEvent, *MIDIRawEventBuffer;

MIDIRawEventBuffer IncomingEventBuf;
MIDIRawEventBuffer NextEventPtr;
Widget		   StopRecordDlg;
Boolean		   Recording;
int		   MidiEventBufferTimer = 0;

extern int seqfd;
extern int midi_dev;

void Midi_SeqStopRecordingCB(Widget w, XtPointer a, XtPointer b)
{
BEGIN("Midi_SeqStopRecordingCB");

    XtDestroyWidget(StopRecordDlg);

    Recording = False;

    /* clear down the widget and close the port */

    close(seqfd);
END;
}
	
void Midi_SeqReadTrack()
{
struct tms ThrowAway;
Widget     StopRecordButton;
XPoint     op;

BEGIN("Midi_SeqReadTrack");

	IncomingEventBuf = (MIDIRawEventBuffer)malloc(MidiEventBufferSize
                                                      * sizeof(MIDIRawEvent));

	if (!IncomingEventBuf)
	{
		Error(NON_FATAL_REPORT_TO_MSGBOX,
                                   "Unable to allocate event buffer.");
		END;
	}

	NextEventPtr = IncomingEventBuf;

        /* Midi_SeqInit pops up its own dialogs for failures */
/*
        if ( !Midi_DeviceInit(O_RDONLY, "/dev/midi00") )
        {
            return;
        }
        if ( ( ( seqfd = open ( "/dev/midi00", O_RDONLY, 0 ) ) ) == -1 )
        {
            fprintf(stderr,"CAN'T open, errno = %d\n",errno);
        }
*/

	StopRecordDlg = XtCreatePopupShell("Stop Recording Dialogue",
                                transientShellWidgetClass, topLevel, NULL, 0);

	StopRecordButton = YCreateSurroundedWidget("STOP", commandWidgetClass,
                                   StopRecordDlg, SurroundShade, NoShade);

	XtAddCallback(StopRecordButton, XtNcallback, Midi_SeqStopRecordingCB,
                                                              NULL);

	op = YPlacePopupAndWarp(StopRecordDlg, XtGrabNonexclusive,
				StopRecordButton, StopRecordButton);

	YAssertDialogueActions(StopRecordDlg, StopRecordButton,
			       StopRecordButton, NULL); 

	Recording = True;

	while(XtAppPending(appContext))
	{
		XtAppProcessEvent(appContext, XtIMAll);
	}

	Recording = True;

	Midi_SeqRecord();

        if ( Recording == True )
            Midi_SeqStopRecordingCB(NULL, NULL, NULL);

END;
}



int
Midi_SeqRecord(void)
{
byte    LastStatus   = 0x90;
byte    NumOperands  = 0;
byte	OperandsLeft = 0;

byte    InByte;
byte   *BytePtr;
byte   LastByte='0';
int out;

BEGIN("Midi_SeqRecord");

    while(Recording)
    {
        /* read the midi port */
        while (XtAppPending(appContext)) XtAppProcessEvent(appContext,
                                                   XtIMAll);
        {
            out = read(seqfd, &InByte, 1);
        }

        if ( out )
        {
        if (IsStatusByte(InByte))
        {
            if (MessageType(InByte) == MIDI_SYSTEM_MSG)
            {
                NumOperands = 0;
            }
            else if (MessageType(InByte) == MIDI_CTRL_CHANGE ||
                       MessageType(InByte) == MIDI_PROG_CHANGE)
            {
                NumOperands = 1;
            }
            else NumOperands = 2;
            LastStatus = InByte;
        }

        if (NumOperands)
        {
            NextEventPtr->Time = NULL;  /* to be arranged */
            NextEventPtr->Bytes[0] = LastStatus;

            OperandsLeft = NumOperands;
            BytePtr = &NextEventPtr->Bytes[1];

            while(OperandsLeft)
            {
                while(!read(seqfd, BytePtr, 1));
                if (!IsStatusByte(*BytePtr))
                {
                    ++BytePtr;
                    --OperandsLeft;
                }
            }

            fprintf(stderr, "%ld %x %x %x\n",
                            NextEventPtr->Time,
                            (int)NextEventPtr->Bytes[0],
                            (int)NextEventPtr->Bytes[1],
                            (int)NextEventPtr->Bytes[2]);
            ++NextEventPtr;
        }

        }
        while (XtAppPending(appContext)) XtAppProcessEvent(appContext, XtIMAll);
    }

    Midi_TrackListAddTrack(Midi_SeqConvertRawDataToEventList());

END;
}


EventList Midi_SeqConvertRawDataToEventList()
{
EventList          NewTrack, RunningPtr, OnePointTrack;
MIDIEventStruct    EventBuffer;
MIDIRawEventBuffer EventPtr;

BEGIN("Midi_SeqConvertRawDataToEventList");

	NewTrack   = NULL;
	RunningPtr = NULL;

	EventPtr = IncomingEventBuf;

	if (MIDIHeaderBuffer.Format == MIDI_NO_FILE_LOADED) 
	{
		Midi_TrackListSetupDefaults();
	}
	while (EventPtr != NextEventPtr)
	{
		EventBuffer.DeltaTime = (((float)(EventPtr->Time * 10000L) / RecordTempo))
					 * MIDIHeaderBuffer.Timing.Division;
		EventBuffer.EventCode = EventPtr->Bytes[0];

		EventBuffer.EventData.NoteOn.Note     = EventPtr->Bytes[1];
		EventBuffer.EventData.NoteOn.Velocity = EventPtr->Bytes[2];

		if (!NewTrack)
		{
			NewTrack = Midi_EventCreateList(&EventBuffer, True);
			RunningPtr = NewTrack;
		}
		else
		{
			RunningPtr = (EventList)Nconc(RunningPtr, Midi_EventCreateList(&EventBuffer, True));
		}
		++EventPtr;
	}

	EventBuffer.EventCode = MIDI_FILE_META_EVENT;
	EventBuffer.EventData.MetaEvent.MetaEventCode = MIDI_END_OF_TRACK;
	EventBuffer.EventData.MetaEvent.NBytes = 0;

	if (!NewTrack)
	{
		NewTrack = Midi_EventCreateList(&EventBuffer, True);
		RunningPtr = NewTrack;
	}
	else
	{
		RunningPtr = (EventList)Nconc(RunningPtr, Midi_EventCreateList(&EventBuffer, True));
	}

	Midi_TrackConvertToOnePointRepresentation(NewTrack);

RETURN_PTR(NewTrack);
}


